module cache_coherence ( input RMS,   // read miss, shared
input RME,                           // read miss, exclusive
input WM,                           // write miss
input WH,                           // write hit
input SHR,                         // snoop hit on read
input SHW,                         // snoop hit on write
input READ_DONE,
input clk,
input reset,
input send_abort,
input write_back_done,
input AllInvDone,
input [2:0] state,
output reg [2:0] new_state,
output reg Cache_Sector_Fill, Invalidate, AdrRetry
);
parameter
INVALID = 3'b000,
SHARED_1 = 3'b001,
EXCLUSIVE = 3'b010,
MODIFIED = 3'b011,
Cache_Fill = 3'b100,
start_write_back = 3'b101,
WaitUntilAllInv = 3'b110;
// Declare current_state and next_state variable.
reg [2:0]  present_state;
reg [2:0]  next_state;
always @(present_state or RMS or RME or WM or WH or SHW or READ_DONE
         or SHR or write_back_done or AllInvDone or send_abort or state or reset)
 begin
Cache_Sector_Fill = 0; // Default values
Invalidate = 0;
AdrRetry = 0;
next_state = present_state;
if (reset)
  next_state = state;
else begin
case(present_state) 
             
INVALID:
// ReadMiss (shared/exclusive), Write Miss
if (RMS || RME || WM)  
begin
Cache_Sector_Fill = 1;
next_state = Cache_Fill ;
end
else
begin
next_state = INVALID;
end

Cache_Fill:
if (send_abort)
begin
next_state = INVALID;
end
else if (READ_DONE)
begin
if (RMS)
begin
next_state = SHARED_1;
end
else if (RME)
begin
next_state = EXCLUSIVE;
end
else if (WM)
begin
Invalidate = 1;
next_state = WaitUntilAllInv;
end
else
begin
next_state = Cache_Fill ;
end
end
else
begin
next_state = Cache_Fill ;
end

SHARED_1:  
if (SHW) // Snoop Hit on a Write.
begin
next_state = INVALID;
end
else if (WH) // Write Hit
begin
Invalidate = 1;
next_state = WaitUntilAllInv;
end
else
begin // Snoop Hit on a Read or Read Hit or any other
next_state = SHARED_1;
end

WaitUntilAllInv:
if (AllInvDone)
begin
next_state = MODIFIED;
end
else
begin
next_state = WaitUntilAllInv;
end

EXCLUSIVE:
if (SHR) // Snoop Hit on a Read:
begin
AdrRetry = 0;
next_state = SHARED_1;  
end
else if (SHW) // Snoop Hit on a Write
begin
next_state = INVALID;
end
else if (WH) // Write Hit
begin
next_state = MODIFIED;
end
else
begin // Read Hit
next_state = EXCLUSIVE;
end

MODIFIED:  
if (SHW) // Snoop Hit on a Write
begin
next_state = INVALID;
end
else if (SHR) // Snoop Hit on a Read
begin
AdrRetry = 1;
next_state = start_write_back;
end
else // Read Hit or Write Hit or anything else.
begin
next_state = MODIFIED;
end
start_write_back:
if (write_back_done)
begin
next_state = SHARED_1;
end
else
begin
next_state = start_write_back;
end
endcase
end
end

/* Sequential */

always @(posedge clk)
begin
present_state = next_state;
new_state = next_state;
end
endmodule




