`define TAG 31:13		// position of tag in address
`define INDEX 12:2		// position of index in address
`define OFFSET 1:0		// position of offset in address

module inst_cache
#(
  // Cache parameters
  parameter SIZE = 16*1024*8,
  parameter NWAYS = 8,
  parameter NSETS = 2048,
  parameter BLOCK_SIZE = 32,
  parameter WIDTH = 32,
  // Memory related paramter, make sure it matches memory module
  parameter MWIDTH = 32,		// same as block size
  // More cache related parameters
  parameter INDEX_WIDTH = 11,
  parameter TAG_WIDTH = 19,
  parameter OFFSET_WIDTH = 2
)
(
  input  wire clock,
  input  wire [WIDTH-1:0] address,     // address form CPU
  input  wire [WIDTH-1:0] din,        // data from CPU (if st inst)
  input  wire rden,                   // 1 if ld instruction
  input  wire wren,                   // 1 if st instruction
  output wire hit_miss,               // 1 if hit, 0 while handling miss
  output wire [WIDTH-1:0] q,          // data from cache to CPU
  output wire [MWIDTH-1:0] mdout,     // data from cache to memory
  output wire [WIDTH-1:0] mrdaddress, // memory read address
  output wire mrden,                  // read enable, 1 if reading from memory
  output wire [WIDTH-1:0] mwraddress, // memory write address
  output wire mwren,                  // write enable, 1 if writing to memory
  input  wire [MWIDTH-1:0] mq         // data coming from memory
);

/***********************
* Global Parameters and Initializations
***********************/

// WAY 1 cache data
reg valid1 [0:NSETS-1];
reg dirty1 [0:NSETS-1];
reg lru1   [0:NSETS-1];
reg [TAG_WIDTH-1:0] tag1 [0:NSETS-1];
reg [MWIDTH-1:0] mem1 [0:NSETS-1] ;
// WAY 2 cache data
reg valid2 [0:NSETS-1];
reg dirty2 [0:NSETS-1];
reg lru2   [0:NSETS-1];
reg [TAG_WIDTH-1:0] tag2 [0:NSETS-1];
reg [MWIDTH-1:0] mem2 [0:NSETS-1] ;

// WAY 3 cache data
reg valid3 [0:NSETS-1];
reg dirty3 [0:NSETS-1];
reg lru3   [0:NSETS-1];
reg [TAG_WIDTH-1:0] tag3 [0:NSETS-1];
reg [MWIDTH-1:0] mem3 [0:NSETS-1] ;

// WAY 4 cache data
reg valid4 [0:NSETS-1];
reg dirty4 [0:NSETS-1];
reg lru4   [0:NSETS-1];
reg [TAG_WIDTH-1:0] tag4 [0:NSETS-1];
reg [MWIDTH-1:0] mem4 [0:NSETS-1];

// WAY 5 cache data
reg valid5 [0:NSETS-1];
reg dirty5 [0:NSETS-1];
reg lru5   [0:NSETS-1];
reg [TAG_WIDTH-1:0] tag5 [0:NSETS-1];
reg [MWIDTH-1:0] mem5 [0:NSETS-1] ;

// WAY 6 cache data
reg valid6 [0:NSETS-1];
reg dirty6 [0:NSETS-1];
reg lru6   [0:NSETS-1];
reg [TAG_WIDTH-1:0] tag6 [0:NSETS-1];
reg [MWIDTH-1:0] mem6 [0:NSETS-1];

// WAY 7 cache data
reg valid7 [0:NSETS-1];
reg dirty7 [0:NSETS-1];
reg lru7   [0:NSETS-1];
reg [TAG_WIDTH-1:0] tag7 [0:NSETS-1];
reg [MWIDTH-1:0] mem7 [0:NSETS-1] ;

// WAY 8 cache data
reg valid8 [0:NSETS-1];
reg dirty8 [0:NSETS-1];
reg lru8   [0:NSETS-1];
reg [TAG_WIDTH-1:0] tag8 [0:NSETS-1];
reg [MWIDTH-1:0] mem8 [0:NSETS-1] ;

// initialize everything to 0
integer k;
initial
begin
	for(k = 0; k < NSETS; k = k +1)
	begin
		valid1[k] = 0;
		valid2[k] = 0;
		valid3[k] = 0;
		valid4[k] = 0;
		valid5[k] = 0;
		valid6[k] = 0;
		valid7[k] = 0;
		valid8[k] = 0;
		dirty1[k] = 0;
		dirty2[k] = 0;
		dirty3[k] = 0;
		dirty4[k] = 0;
		dirty5[k] = 0;
		dirty6[k] = 0;
		dirty7[k] = 0;
		dirty8[k] = 0;
		lru1[k] = 0;
		lru2[k] = 0;
		lru3[k] = 0;
		lru4[k] = 0;
		lru5[k] = 0;
		lru6[k] = 0;
		lru7[k] = 0;
		lru8[k] = 0;
	end
end

// internal registers
reg _hit_miss = 1'b0;
reg [WIDTH-1:0]	_q = {WIDTH{1'b0}};
reg [MWIDTH-1:0] _mdout = {MWIDTH{1'b0}};
reg [WIDTH-1:0]	_mwraddress = {WIDTH{1'b0}};
reg _mwren = 1'b0;

// output assignments of internal registers
assign hit_miss = _hit_miss;
assign mrden = !((valid1[address[`INDEX]] && (tag1[address[`INDEX]] == address[`TAG]))
			  || (valid2[address[`INDEX]] && (tag2[address[`INDEX]] == address[`TAG]))
			  || (valid3[address[`INDEX]] && (tag3[address[`INDEX]] == address[`TAG]))
			  || (valid4[address[`INDEX]] && (tag4[address[`INDEX]] == address[`TAG]))
			  || (valid5[address[`INDEX]] && (tag5[address[`INDEX]] == address[`TAG]))
			  || (valid6[address[`INDEX]] && (tag6[address[`INDEX]] == address[`TAG]))
			  || (valid7[address[`INDEX]] && (tag7[address[`INDEX]] == address[`TAG]))
			  || (valid8[address[`INDEX]] && (tag8[address[`INDEX]] == address[`TAG])));
assign mwren = _mwren;
assign mdout = _mdout;
assign mrdaddress = {address[`TAG],address[`INDEX]};
assign mwraddress = _mwraddress;
assign q = _q;

// state parameters
localparam idle = 1'b0;		// receive requests from CPU 
localparam miss = 1'b1;	// miss state: write back dirty block and request memory data

// state register
reg currentState = idle;

/***********************
* State Machine
***********************/

always @(posedge clock)
begin
	case (currentState)
		idle: begin
			// reset write enable, if it was turned on
			_mwren <= 0;
			// set _hit_miss register
			_hit_miss <= ((valid1[address[`INDEX]] && (tag1[address[`INDEX]] == address[`TAG]))
					  ||  (valid2[address[`INDEX]] && (tag2[address[`INDEX]] == address[`TAG]))
					  ||  (valid3[address[`INDEX]] && (tag3[address[`INDEX]] == address[`TAG]))
					  ||  (valid4[address[`INDEX]] && (tag4[address[`INDEX]] == address[`TAG]))
					  ||  (valid5[address[`INDEX]] && (tag5[address[`INDEX]] == address[`TAG]))
					  ||  (valid6[address[`INDEX]] && (tag6[address[`INDEX]] == address[`TAG]))
					  ||  (valid7[address[`INDEX]] && (tag7[address[`INDEX]] == address[`TAG]))
					  ||  (valid8[address[`INDEX]] && (tag8[address[`INDEX]] == address[`TAG])));

			// do nothing on null request
			if(~rden && ~wren) currentState <= idle;
			
			// check way 1
			else if(valid1[address[`INDEX]] && (tag1[address[`INDEX]] == address[`TAG]))
			begin
				// read hit
				if(rden) _q <= mem1[address[`INDEX]];
				// write hit
				else if(wren)
				begin
					_q = {WIDTH{1'b0}};
					mem1[address[`INDEX]] <= din;
					dirty1[address[`INDEX]] <= 1;
				end
				// update LRU data
				lru1[address[`INDEX]] <= 0;
				if(lru2[address[`INDEX]] <= lru1[address[`INDEX]]) lru2[address[`INDEX]] <= lru2[address[`INDEX]] + 1;
				if(lru3[address[`INDEX]] <= lru1[address[`INDEX]]) lru3[address[`INDEX]] <= lru3[address[`INDEX]] + 1;
				if(lru4[address[`INDEX]] <= lru1[address[`INDEX]]) lru4[address[`INDEX]] <= lru4[address[`INDEX]] + 1;
				if(lru5[address[`INDEX]] <= lru1[address[`INDEX]]) lru5[address[`INDEX]] <= lru5[address[`INDEX]] + 1;
				if(lru6[address[`INDEX]] <= lru1[address[`INDEX]]) lru6[address[`INDEX]] <= lru6[address[`INDEX]] + 1;
				if(lru7[address[`INDEX]] <= lru1[address[`INDEX]]) lru7[address[`INDEX]] <= lru7[address[`INDEX]] + 1;
				if(lru8[address[`INDEX]] <= lru1[address[`INDEX]]) lru8[address[`INDEX]] <= lru8[address[`INDEX]] + 1;
			end
			
			// check way 2
			else if(valid2[address[`INDEX]] && (tag2[address[`INDEX]] == address[`TAG]))
			begin
				// read hit
				if(rden) _q <= mem2[address[`INDEX]];
				// write hit
				else if(wren)
				begin
					_q = {WIDTH{1'b0}};
					mem2[address[`INDEX]] <= din;
					dirty2[address[`INDEX]] <= 1;
				end
				// update LRU data
				if(lru1[address[`INDEX]] <= lru2[address[`INDEX]]) lru1[address[`INDEX]] <= lru1[address[`INDEX]] + 1;
				if(lru3[address[`INDEX]] <= lru2[address[`INDEX]]) lru3[address[`INDEX]] <= lru3[address[`INDEX]] + 1;
				if(lru4[address[`INDEX]] <= lru2[address[`INDEX]]) lru4[address[`INDEX]] <= lru4[address[`INDEX]] + 1;
				if(lru5[address[`INDEX]] <= lru2[address[`INDEX]]) lru5[address[`INDEX]] <= lru5[address[`INDEX]] + 1;
				if(lru6[address[`INDEX]] <= lru2[address[`INDEX]]) lru6[address[`INDEX]] <= lru6[address[`INDEX]] + 1;
				if(lru7[address[`INDEX]] <= lru2[address[`INDEX]]) lru7[address[`INDEX]] <= lru7[address[`INDEX]] + 1;
				if(lru8[address[`INDEX]] <= lru2[address[`INDEX]]) lru8[address[`INDEX]] <= lru8[address[`INDEX]] + 1;
				lru2[address[`INDEX]] <= 0;
                                end

			// check way 3
			else if(valid3[address[`INDEX]] && (tag3[address[`INDEX]] == address[`TAG]))
			begin
				// read hit
				if(rden) _q <= mem3[address[`INDEX]];
				// write hit
				else if(wren)
				begin
					_q = {WIDTH{1'b0}};
					mem3[address[`INDEX]] <= din;
					dirty3[address[`INDEX]] <= 1;
				end
				// update LRU data
				if(lru1[address[`INDEX]] <= lru3[address[`INDEX]]) lru1[address[`INDEX]] <= lru1[address[`INDEX]] + 1;
				if(lru2[address[`INDEX]] <= lru3[address[`INDEX]]) lru2[address[`INDEX]] <= lru2[address[`INDEX]] + 1;
				if(lru4[address[`INDEX]] <= lru3[address[`INDEX]]) lru4[address[`INDEX]] <= lru4[address[`INDEX]] + 1;
				if(lru5[address[`INDEX]] <= lru3[address[`INDEX]]) lru5[address[`INDEX]] <= lru5[address[`INDEX]] + 1;
				if(lru6[address[`INDEX]] <= lru3[address[`INDEX]]) lru6[address[`INDEX]] <= lru6[address[`INDEX]] + 1;
				if(lru7[address[`INDEX]] <= lru3[address[`INDEX]]) lru7[address[`INDEX]] <= lru7[address[`INDEX]] + 1;
				if(lru8[address[`INDEX]] <= lru3[address[`INDEX]]) lru8[address[`INDEX]] <= lru8[address[`INDEX]] + 1;
				lru3[address[`INDEX]] <= 0;
                          end

			// check way 4
			else if(valid4[address[`INDEX]] && (tag4[address[`INDEX]] == address[`TAG]))
			begin
				// read hit
				if(rden) _q <= mem4[address[`INDEX]];
				// write hit
				else if(wren)
				begin
					_q = {WIDTH{1'b0}};
					mem4[address[`INDEX]] <= din;
					dirty4[address[`INDEX]] <= 1;
				end
				// update LRU data
				if(lru1[address[`INDEX]] <= lru4[address[`INDEX]]) lru1[address[`INDEX]] <= lru1[address[`INDEX]] + 1;
				if(lru2[address[`INDEX]] <= lru4[address[`INDEX]]) lru2[address[`INDEX]] <= lru2[address[`INDEX]] + 1;
				if(lru3[address[`INDEX]] <= lru4[address[`INDEX]]) lru3[address[`INDEX]] <= lru3[address[`INDEX]] + 1;
				if(lru5[address[`INDEX]] <= lru4[address[`INDEX]]) lru5[address[`INDEX]] <= lru5[address[`INDEX]] + 1;
				if(lru6[address[`INDEX]] <= lru4[address[`INDEX]]) lru6[address[`INDEX]] <= lru6[address[`INDEX]] + 1;
				if(lru7[address[`INDEX]] <= lru4[address[`INDEX]]) lru7[address[`INDEX]] <= lru7[address[`INDEX]] + 1;
				if(lru8[address[`INDEX]] <= lru4[address[`INDEX]]) lru8[address[`INDEX]] <= lru8[address[`INDEX]] + 1;
				lru4[address[`INDEX]] <= 0;
                        end

			// check way 5
			else if(valid5[address[`INDEX]] && (tag5[address[`INDEX]] == address[`TAG]))
			begin
				// read hit
				if(rden) _q <= mem5[address[`INDEX]];
				// write hit
				else if(wren)
				begin
					_q = {WIDTH{1'b0}};
					mem5[address[`INDEX]] <= din;
					dirty5[address[`INDEX]] <= 1;
				end
				// update LRU data
				if(lru1[address[`INDEX]] <= lru5[address[`INDEX]]) lru1[address[`INDEX]] <= lru1[address[`INDEX]] + 1;
				if(lru2[address[`INDEX]] <= lru5[address[`INDEX]]) lru2[address[`INDEX]] <= lru2[address[`INDEX]] + 1;
				if(lru3[address[`INDEX]] <= lru5[address[`INDEX]]) lru3[address[`INDEX]] <= lru3[address[`INDEX]] + 1;
				if(lru4[address[`INDEX]] <= lru5[address[`INDEX]]) lru4[address[`INDEX]] <= lru4[address[`INDEX]] + 1;
				if(lru6[address[`INDEX]] <= lru5[address[`INDEX]]) lru6[address[`INDEX]] <= lru6[address[`INDEX]] + 1;
				if(lru7[address[`INDEX]] <= lru5[address[`INDEX]]) lru7[address[`INDEX]] <= lru7[address[`INDEX]] + 1;
				if(lru8[address[`INDEX]] <= lru5[address[`INDEX]]) lru8[address[`INDEX]] <= lru8[address[`INDEX]] + 1;
				lru5[address[`INDEX]] <= 0;
                         end

			// check way 6
			else if(valid6[address[`INDEX]] && (tag6[address[`INDEX]] == address[`TAG]))
			begin
				// read hit
				if(rden) _q <= mem6[address[`INDEX]];
				// write hit
				else if(wren)
				begin
					_q = {WIDTH{1'b0}};
					mem6[address[`INDEX]] <= din;
					dirty6[address[`INDEX]] <= 1;
				end
				// update LRU data
				if(lru1[address[`INDEX]] <= lru6[address[`INDEX]]) lru1[address[`INDEX]] <= lru1[address[`INDEX]] + 1;
				if(lru2[address[`INDEX]] <= lru6[address[`INDEX]]) lru2[address[`INDEX]] <= lru2[address[`INDEX]] + 1;
				if(lru3[address[`INDEX]] <= lru6[address[`INDEX]]) lru3[address[`INDEX]] <= lru3[address[`INDEX]] + 1;
				if(lru4[address[`INDEX]] <= lru6[address[`INDEX]]) lru4[address[`INDEX]] <= lru4[address[`INDEX]] + 1;
				if(lru5[address[`INDEX]] <= lru6[address[`INDEX]]) lru5[address[`INDEX]] <= lru5[address[`INDEX]] + 1;
				if(lru7[address[`INDEX]] <= lru6[address[`INDEX]]) lru7[address[`INDEX]] <= lru7[address[`INDEX]] + 1;
				if(lru8[address[`INDEX]] <= lru6[address[`INDEX]]) lru8[address[`INDEX]] <= lru8[address[`INDEX]] + 1;
				lru6[address[`INDEX]] <= 0;
                         end

			// check way 7
			else if(valid7[address[`INDEX]] && (tag7[address[`INDEX]] == address[`TAG]))
			begin
				// read hit
				if(rden) _q <= mem7[address[`INDEX]];
				// write hit
				else if(wren)
				begin
					_q = {WIDTH{1'b0}};
					mem7[address[`INDEX]] <= din;
					dirty7[address[`INDEX]] <= 1;
				end
				// update LRU data
				if(lru1[address[`INDEX]] <= lru7[address[`INDEX]]) lru1[address[`INDEX]] <= lru1[address[`INDEX]] + 1;
				if(lru2[address[`INDEX]] <= lru7[address[`INDEX]]) lru2[address[`INDEX]] <= lru2[address[`INDEX]] + 1;
				if(lru3[address[`INDEX]] <= lru7[address[`INDEX]]) lru3[address[`INDEX]] <= lru3[address[`INDEX]] + 1;
				if(lru4[address[`INDEX]] <= lru7[address[`INDEX]]) lru4[address[`INDEX]] <= lru4[address[`INDEX]] + 1;
				if(lru5[address[`INDEX]] <= lru7[address[`INDEX]]) lru5[address[`INDEX]] <= lru5[address[`INDEX]] + 1;
				if(lru6[address[`INDEX]] <= lru7[address[`INDEX]]) lru6[address[`INDEX]] <= lru6[address[`INDEX]] + 1;
				if(lru8[address[`INDEX]] <= lru7[address[`INDEX]]) lru8[address[`INDEX]] <= lru8[address[`INDEX]] + 1;
				lru7[address[`INDEX]] <= 0;
                         end

			// check way 8
			else if(valid8[address[`INDEX]] && (tag8[address[`INDEX]] == address[`TAG]))
			begin
				// read hit
				if(rden) _q <= mem8[address[`INDEX]];
				// write hit
				else if(wren)
				begin
					_q = {WIDTH{1'b0}};
					mem8[address[`INDEX]] <= din;
					dirty8[address[`INDEX]] <= 1;
				end
				// update LRU data
				if(lru1[address[`INDEX]] <= lru8[address[`INDEX]]) lru1[address[`INDEX]] <= lru1[address[`INDEX]] + 1;
				if(lru2[address[`INDEX]] <= lru8[address[`INDEX]]) lru2[address[`INDEX]] <= lru2[address[`INDEX]] + 1;
				if(lru3[address[`INDEX]] <= lru8[address[`INDEX]]) lru3[address[`INDEX]] <= lru3[address[`INDEX]] + 1;
				if(lru4[address[`INDEX]] <= lru8[address[`INDEX]]) lru4[address[`INDEX]] <= lru4[address[`INDEX]] + 1;
				if(lru5[address[`INDEX]] <= lru8[address[`INDEX]]) lru5[address[`INDEX]] <= lru5[address[`INDEX]] + 1;
				if(lru6[address[`INDEX]] <= lru8[address[`INDEX]]) lru6[address[`INDEX]] <= lru6[address[`INDEX]] + 1;
				if(lru7[address[`INDEX]] <= lru8[address[`INDEX]]) lru7[address[`INDEX]] <= lru7[address[`INDEX]] + 1;
				lru8[address[`INDEX]] <= 0;
			end
			
			// miss
			else currentState <= miss;
		end
	
		miss: begin
			// one of the ways is invalid -- no need to evict
			if(~valid1[address[`INDEX]])
			begin
				mem1[address[`INDEX]] <= mq;
				tag1[address[`INDEX]] <= address[`TAG];
				dirty1[address[`INDEX]] <= 0;
				valid1[address[`INDEX]] <= 1;
			end
			
			else if(~valid2[address[`INDEX]])
			begin
				mem2[address[`INDEX]] <= mq;
				tag2[address[`INDEX]] <= address[`TAG];
				dirty2[address[`INDEX]] <= 0;
				valid2[address[`INDEX]] <= 1;
			end
			else if(~valid3[address[`INDEX]])
			begin
				mem3[address[`INDEX]] <= mq;
				tag3[address[`INDEX]] <= address[`TAG];
				dirty3[address[`INDEX]] <= 0;
				valid3[address[`INDEX]] <= 1;
			end

			else if(~valid4[address[`INDEX]])
			begin
				mem4[address[`INDEX]] <= mq;
				tag4[address[`INDEX]] <= address[`TAG];
				dirty4[address[`INDEX]] <= 0;
				valid4[address[`INDEX]] <= 1;
			end


			else if(~valid5[address[`INDEX]])
			begin
				mem5[address[`INDEX]] <= mq;
				tag5[address[`INDEX]] <= address[`TAG];
				dirty5[address[`INDEX]] <= 0;
				valid5[address[`INDEX]] <= 1;
			end


			else if(~valid6[address[`INDEX]])
			begin
				mem6[address[`INDEX]] <= mq;
				tag6[address[`INDEX]] <= address[`TAG];
				dirty6[address[`INDEX]] <= 0;
				valid6[address[`INDEX]] <= 1;
			end


			else if(~valid7[address[`INDEX]])
			begin
				mem7[address[`INDEX]] <= mq;
				tag7[address[`INDEX]] <= address[`TAG];
				dirty7[address[`INDEX]] <= 0;
				valid7[address[`INDEX]] <= 1;
			end


			else if(~valid8[address[`INDEX]])
			begin
				mem8[address[`INDEX]] <= mq;
				tag8[address[`INDEX]] <= address[`TAG];
				dirty8[address[`INDEX]] <= 0;
				valid8[address[`INDEX]] <= 1;
			end

			
			// way 1 is LRU
			else if(lru1[address[`INDEX]] == 1)
			begin
				// dirty block writeback
				if(dirty1[address[`INDEX]] == 1)
				begin
					_mwraddress <= {tag1[address[`INDEX]],address[`INDEX]}; 
					_mwren <= 1;
					_mdout <= mem1[address[`INDEX]];
				end
				mem1[address[`INDEX]] <= mq;
				tag1[address[`INDEX]] <= address[`TAG];
				dirty1[address[`INDEX]] <= 0;
				valid1[address[`INDEX]] <= 1;
			end
			
			// way 2 is LRU
			else if(lru2[address[`INDEX]] == 1)
			begin
				// dirty block writeback
				if(dirty2[address[`INDEX]] == 1)
				begin
					_mwraddress <= {tag2[address[`INDEX]],address[`INDEX]}; 
					_mwren <= 1;
					_mdout <= mem2[address[`INDEX]];
				end
				mem2[address[`INDEX]] <= mq;
				tag2[address[`INDEX]] <= address[`TAG];
				dirty2[address[`INDEX]] <= 0;
				valid2[address[`INDEX]] <= 1;
			end

			// way 3 is LRU
			else if(lru3[address[`INDEX]] == 1)
			begin
				// dirty block writeback
				if(dirty3[address[`INDEX]] == 1)
				begin
					_mwraddress <= {tag3[address[`INDEX]],address[`INDEX]}; 
					_mwren <= 1;
					_mdout <= mem3[address[`INDEX]];
				end
				mem3[address[`INDEX]] <= mq;
				tag3[address[`INDEX]] <= address[`TAG];
				dirty3[address[`INDEX]] <= 0;
				valid3[address[`INDEX]] <= 1;
			end

			// way 4 is LRU
			else if(lru4[address[`INDEX]] == 1)
			begin
				// dirty block writeback
				if(dirty4[address[`INDEX]] == 1)
				begin
					_mwraddress <= {tag4[address[`INDEX]],address[`INDEX]}; 
					_mwren <= 1;
					_mdout <= mem4[address[`INDEX]];
				end
				mem4[address[`INDEX]] <= mq;
				tag4[address[`INDEX]] <= address[`TAG];
				dirty4[address[`INDEX]] <= 0;
				valid4[address[`INDEX]] <= 1;
			end

			// way 5 is LRU
			else if(lru5[address[`INDEX]] == 1)
			begin
				// dirty block writeback
				if(dirty5[address[`INDEX]] == 1)
				begin
					_mwraddress <= {tag5[address[`INDEX]],address[`INDEX]}; 
					_mwren <= 1;
					_mdout <= mem5[address[`INDEX]];
				end
				mem5[address[`INDEX]] <= mq;
				tag5[address[`INDEX]] <= address[`TAG];
				dirty5[address[`INDEX]] <= 0;
				valid5[address[`INDEX]] <= 1;
			end

			// way 6 is LRU
			else if(lru6[address[`INDEX]] == 1)
			begin
				// dirty block writeback
				if(dirty6[address[`INDEX]] == 1)
				begin
					_mwraddress <= {tag6[address[`INDEX]],address[`INDEX]}; 
					_mwren <= 1;
					_mdout <= mem6[address[`INDEX]];
				end
				mem6[address[`INDEX]] <= mq;
				tag6[address[`INDEX]] <= address[`TAG];
				dirty6[address[`INDEX]] <= 0;
				valid6[address[`INDEX]] <= 1;
			end

			// way 7 is LRU
			else if(lru7[address[`INDEX]] == 1)
			begin
				// dirty block writeback
				if(dirty7[address[`INDEX]] == 1)
				begin
					_mwraddress <= {tag7[address[`INDEX]],address[`INDEX]}; 
					_mwren <= 1;
					_mdout <= mem7[address[`INDEX]];
				end
				mem7[address[`INDEX]] <= mq;
				tag7[address[`INDEX]] <= address[`TAG];
				dirty7[address[`INDEX]] <= 0;
				valid7[address[`INDEX]] <= 1;
			end

			// way 8 is LRU
			else if(lru8[address[`INDEX]] == 1)
			begin
				// dirty block writeback
				if(dirty8[address[`INDEX]] == 1)
				begin
					_mwraddress <= {tag8[address[`INDEX]],address[`INDEX]}; 
					_mwren <= 1;
					_mdout <= mem8[address[`INDEX]];
				end
				mem8[address[`INDEX]] <= mq;
				tag8[address[`INDEX]] <= address[`TAG];
				dirty8[address[`INDEX]] <= 0;
				valid8[address[`INDEX]] <= 1;
			end

			currentState <= idle;
		end

		default: currentState <= idle;

	endcase

end

endmodule